import time
from typing import Protocol, Tuple

class MaskLike(Protocol):
    limit: int
    def __getitem__(self, idx: int) -> int: ...

def sieve_odd_mask(limit: int) -> bytearray:
    if limit < 2:
        return bytearray(b"\x00")
    size = (limit // 2) + 1
    mask = bytearray(b"\x01") * size
    mask[0] = 0
    p = 3
    while p * p <= limit:
        if mask[p >> 1]:
            for x in range(p * p, limit + 1, 2 * p):
                mask[x >> 1] = 0
        p += 2
    return mask

def autotune_chunk(mask_bytes: int, cpu_features: set[str] | None = None) -> int:
    feats = cpu_features or set()
    if mask_bytes < 8_000_000:
        base = 16384
    elif mask_bytes < 64_000_000:
        base = 32768
    elif mask_bytes < 256_000_000:
        base = 65536
    else:
        base = 131072
    if "avx512f" in feats:
        base *= 2
    elif "avx2" in feats:
        base = int(base * 1.5)
    return min(max((base // 1024) * 1024, 8192), 262144)

def count_primes_for_duration(
    mask: MaskLike, duration: float, start_n: int = 1, chunk: int = 32768
) -> Tuple[int, int]:
    end = time.perf_counter() + duration
    n = start_n
    count = 0
    if n <= 2:
        count += 1
        n = 3
    if n % 2 == 0:
        n += 1
    while True:
        stop = n + 2 * chunk
        while n < stop:
            if n > mask.limit:
                return count, n
            count += mask[n >> 1]
            n += 2
        if time.perf_counter() >= end:
            return count, n



from __future__ import annotations

import math
import os
import random
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Tuple

from .mask import cache_dir, open_mask, sieve_to_cache

DEFAULT_LIMIT = 100_000_000
_mask = None  # global singleton MaskView


def _ensure_mask(limit: int):
    global _mask
    if _mask is not None:
        if limit <= _mask.limit:
            return _mask
        _mask.close()
        _mask = None

    path = sieve_to_cache(limit)
    _mask = open_mask(path)
    return _mask


def _ensure_for_n(n: int, *, min_limit: Optional[int] = None):
    need = max(n, min_limit or 0, DEFAULT_LIMIT if min_limit is None else 0)
    if need < 2:
        need = 2
    return _ensure_mask(need)


def isprime(n: int, *, limit: Optional[int] = None) -> bool:
    if n < 2:
        return False
    if n == 2:
        return True
    if (n & 1) == 0:
        return False
    mv = _ensure_mask(max(n, limit or 0, DEFAULT_LIMIT if limit is None else 0))
    return bool(mv[n >> 1])


def isprime_many(nums: Iterable[int], *, limit: Optional[int] = None) -> List[bool]:
    """
    Bulk primality checks. Opens/grows mask once and does O(1) lookups.
    """
    xs = list(nums)
    if not xs:
        return []
    mx = max(xs)
    mv = _ensure_mask(max(mx, limit or 0, DEFAULT_LIMIT if limit is None else 0))

    out: List[bool] = []
    for n in xs:
        if n < 2:
            out.append(False)
        elif n == 2:
            out.append(True)
        elif (n & 1) == 0:
            out.append(False)
        elif n > mv.limit:
            # auto-grow if needed (rare if you passed a reasonable limit)
            mv = _ensure_mask(max(n, mv.limit * 2))
            out.append(bool(mv[n >> 1]))
        else:
            out.append(bool(mv[n >> 1]))
    return out


def primerange(a: int, b: int, *, limit: Optional[int] = None) -> List[int]:
    if b <= 2 or b <= a:
        return []
    hi = b - 1
    mv = _ensure_mask(max(hi, limit or 0, DEFAULT_LIMIT if limit is None else 0))

    out: List[int] = []
    if a <= 2 < b:
        out.append(2)

    start = max(a, 3)
    if (start & 1) == 0:
        start += 1

    # scan odds
    for n in range(start, b, 2):
        if mv[n >> 1]:
            out.append(n)
    return out


def randprime(a: int, b: int, *, limit: Optional[int] = None, tries: int = 64) -> int:
    if b <= a:
        raise ValueError("randprime requires b > a")
    if b <= 2:
        raise ValueError("No primes in the given range")

    hi = b - 1
    mv = _ensure_mask(max(hi, limit or 0, DEFAULT_LIMIT if limit is None else 0))
    lo = max(a, 2)

    for _ in range(max(1, tries)):
        x = random.randrange(lo, b)
        if x == 2:
            return 2
        if x < 2 or (x & 1) == 0:
            continue
        if x > mv.limit:
            mv = _ensure_mask(max(x, mv.limit * 2))
        if mv[x >> 1]:
            return x

    primes = primerange(a, b, limit=mv.limit)
    if not primes:
        raise ValueError("No primes exist in the given range")
    return random.choice(primes)


def primepi(n: int, *, limit: Optional[int] = None) -> int:
    """
    Count primes <= n.
    Uses fast bit_count-based counting on the bitset.
    """
    if n < 2:
        return 0
    mv = _ensure_mask(max(n, limit or 0, DEFAULT_LIMIT if limit is None else 0))

    total = 1  # prime 2
    last_odd = n if (n & 1) else (n - 1)
    if last_odd >= 3:
        idx = last_odd >> 1
        # count ones in [1..idx] => count_ones_upto(idx) - bit0
        total += mv.count_ones_upto(idx) - mv.count_ones_upto(0)
    return total


def _nth_prime_upper(n: int) -> int:
    if n <= 1:
        return 2
    if n < 6:
        return 15
    nn = float(n)
    return int(nn * (math.log(nn) + math.log(math.log(nn))) + 10)


def prime(nth: int, *, limit: Optional[int] = None) -> int:
    """
    Return the nth prime (1-indexed). Uses bit scanning to avoid per-bit Python loops.
    """
    if nth < 1:
        raise ValueError("prime(nth) requires nth >= 1")
    if nth == 1:
        return 2

    target = _nth_prime_upper(nth)
    if limit is not None:
        target = max(target, limit)

    mv = _ensure_mask(max(target, DEFAULT_LIMIT))

    while True:
        count = 1  # prime 2

        # iterate set bits for odd primes (i>=1 -> 3)
        for bit_idx in mv.iter_set_bits(1, mv.nbits):
            count += 1
            if count == nth:
                return (bit_idx << 1) + 1

        mv = _ensure_mask(mv.limit * 2)


def prevprime(n: int, *, limit: Optional[int] = None) -> int:
    if n <= 2:
        raise ValueError("No prime exists below 2")

    x = n - 1
    if x == 2:
        return 2
    if (x & 1) == 0:
        x -= 1

    mv = _ensure_mask(max(x, limit or 0, DEFAULT_LIMIT if limit is None else 0))

    while x >= 3:
        if mv[x >> 1]:
            return x
        x -= 2
    return 2


def nextprime(n: int, *, limit: Optional[int] = None) -> int:
    if n < 2:
        return 2

    x = n + 1
    if x <= 2:
        return 2
    if (x & 1) == 0:
        x += 1

    mv = _ensure_mask(max(x, limit or 0, DEFAULT_LIMIT if limit is None else 0))

    while True:
        if x <= mv.limit:
            if mv[x >> 1]:
                return x
            x += 2
        else:
            mv = _ensure_mask(max(x, mv.limit * 2))


def primes_up_to(n: int, *, limit: Optional[int] = None) -> List[int]:
    if n < 2:
        return []
    return primerange(2, n + 1, limit=limit)


def primes_upto(n: int, *, limit: Optional[int] = None) -> List[int]:
    return primes_up_to(n, limit=limit)


def primes(count: int, *, limit: Optional[int] = None) -> List[int]:
    """
    Return first `count` primes. Uses bit scanning for speed.
    """
    if count < 0:
        raise ValueError("primes(count) requires count >= 0")
    if count == 0:
        return []
    if count == 1:
        return [2]

    target = _nth_prime_upper(count)
    if limit is not None:
        target = max(target, limit)

    mv = _ensure_mask(max(target, DEFAULT_LIMIT))

    while True:
        out = [2]
        for bit_idx in mv.iter_set_bits(1, mv.nbits):
            out.append((bit_idx << 1) + 1)
            if len(out) == count:
                return out
        mv = _ensure_mask(mv.limit * 2)


def factorint(n: int, *, limit: Optional[int] = None) -> Dict[int, int]:
    if n == 0:
        raise ValueError("factorint(0) is undefined")

    factors: Dict[int, int] = {}

    if n < 0:
        factors[-1] = 1
        n = -n
    if n == 1:
        return factors

    # factor 2
    e = 0
    while (n & 1) == 0:
        n >>= 1
        e += 1
    if e:
        factors[2] = e
    if n == 1:
        return factors

    r = math.isqrt(n)
    mv = _ensure_mask(max(r, limit or 0, DEFAULT_LIMIT if limit is None else 0))

    p = 3
    while p <= r and n > 1:
        if mv[p >> 1]:
            if n % p == 0:
                e = 0
                while n % p == 0:
                    n //= p
                    e += 1
                factors[p] = e
                r = math.isqrt(n)
        p += 2

    if n > 1:
        factors[n] = factors.get(n, 0) + 1
    return factors


@dataclass(frozen=True)
class CacheInfo:
    cache_dir: str
    masks: List[Tuple[str, int]]
    current_mask_limit: Optional[int]
    default_limit: int


def cache_info() -> CacheInfo:
    cdir = cache_dir()
    masks: List[Tuple[str, int]] = []

    if cdir.exists():
        for p in sorted(cdir.glob("lookprime_mask_*.lpm")):
            lim = -1
            try:
                lim = int(p.stem.split("_")[-1])
            except Exception:
                lim = -1
            masks.append((str(p), lim))

    current = getattr(_mask, "limit", None) if _mask is not None else None
    return CacheInfo(
        cache_dir=str(cdir),
        masks=masks,
        current_mask_limit=current,
        default_limit=DEFAULT_LIMIT,
    )


def clear_cache(*, close_open_mask: bool = True) -> int:
    global _mask
    if close_open_mask and _mask is not None:
        try:
            _mask.close()
        finally:
            _mask = None

    cdir = cache_dir()
    deleted = 0
    if not cdir.exists():
        return 0

    for p in cdir.glob("lookprime_mask_*.lpm"):
        try:
            p.unlink()
            deleted += 1
        except Exception:
            pass
    return deleted



import platform

def detect_cpu_features() -> set[str]:
    if platform.system() == "Linux":
        try:
            with open("/proc/cpuinfo") as f:
                for l in f:
                    if l.startswith("flags"):
                        return set(l.split(":")[1].split())
        except:
            pass
    return set()

def cpu_summary() -> dict:
    return {
        "os": platform.platform(),
        "arch": platform.machine(),
        "python": platform.python_version(),
        "simd": ", ".join(sorted(detect_cpu_features())) or "unknown"
    }



from .core import sieve_odd_mask, count_primes_for_duration, autotune_chunk
from .mask import (
    MaskView,
    cache_dir,
    default_mask_path,
    open_mask,
    sieve_to_cache,
)
from .cpu import cpu_summary, detect_cpu_features
from .api import (
    isprime,
    isprime_many,
    primerange,
    randprime,
    primepi,
    prime,
    prevprime,
    nextprime,
    primes_up_to,
    primes_upto,
    primes,
    factorint,
    cache_info,
    clear_cache,
    CacheInfo,
)

__all__ = [
    "sieve_odd_mask",
    "count_primes_for_duration",
    "autotune_chunk",
    "MaskView",
    "cache_dir",
    "default_mask_path",
    "open_mask",
    "sieve_to_cache",
    "detect_cpu_features",
    "cpu_summary",
    "isprime",
    "isprime_many",
    "primerange",
    "randprime",
    "primepi",
    "prime",
    "prevprime",
    "nextprime",
    "primes_up_to",
    "primes_upto",
    "primes",
    "factorint",
    "cache_info",
    "clear_cache",
    "CacheInfo",
]



from __future__ import annotations

import argparse
import math
import os
import statistics
import time
from pathlib import Path

from .core import count_primes_for_duration, autotune_chunk
from .cpu import cpu_summary, detect_cpu_features
from .mask import cache_dir, default_mask_path, open_mask, sieve_to_cache
from .api import cache_info, clear_cache, isprime, factorint, DEFAULT_LIMIT


def cmd_info(args: argparse.Namespace) -> int:
    print("System")
    print("------")
    info = cpu_summary()
    for k, v in info.items():
        print(f"{k:>14}: {v}")
    print()

    ci = cache_info()
    print("Cache")
    print("-----")
    print(f"{'cache_dir':>14}: {ci.cache_dir}")
    print(f"{'default_limit':>14}: {ci.default_limit:,}")
    print(f"{'open_mask':>14}: {ci.current_mask_limit:,}" if ci.current_mask_limit else f"{'open_mask':>14}: (none)")
    print()

    print("Masks")
    print("-----")
    if not ci.masks:
        print("(none)")
        return 0

    for path, lim in ci.masks:
        lim_s = f"{lim:,}" if lim and lim > 0 else "?"
        print(f"- limit={lim_s}  file={path}")
    return 0


def cmd_clear_cache(args: argparse.Namespace) -> int:
    deleted = clear_cache()
    print(f"Deleted {deleted} cached mask file(s).")
    return 0


def cmd_isprime(args: argparse.Namespace) -> int:
    n = int(args.n)
    lim = int(args.limit) if args.limit is not None else None
    print(isprime(n, limit=lim))
    return 0


def cmd_factor(args: argparse.Namespace) -> int:
    n = int(args.n)
    lim = int(args.limit) if args.limit is not None else None
    print(factorint(n, limit=lim))
    return 0


def cmd_build(args: argparse.Namespace) -> int:
    limit = int(args.limit)
    out: Path

    if args.output:
        out = Path(args.output)
    else:
        out = default_mask_path(limit)

    if not args.force and out.exists():
        print(f"Mask already exists: {out}")
        return 0

    print(f"Cache dir: {cache_dir()}")
    print(f"Building mask LIMIT={limit:,} ...")
    t0 = time.perf_counter()

    if not args.output:
        path = sieve_to_cache(limit, force=args.force)
        t1 = time.perf_counter()
        print(f"Saved: {path}")
        print(f"Build time: {t1 - t0:.3f}s")
        return 0

    # Build into cache (bitset format) then move to requested output path
    path = sieve_to_cache(limit, force=True)
    Path(path).replace(out)
    t1 = time.perf_counter()
    print(f"Saved: {out.resolve()}")
    print(f"Build time: {t1 - t0:.3f}s")
    return 0


def cmd_benchmark(args: argparse.Namespace) -> int:
    limit = int(args.limit)
    duration = float(args.duration)
    iterations = int(args.iterations)

    print("System")
    print("------")
    info = cpu_summary()
    for k, v in info.items():
        print(f"{k:>10}: {v}")
    print()

    # choose mask path
    if args.mask:
        path = Path(args.mask)
        if not path.exists():
            raise SystemExit(f"Mask file not found: {path}")
    else:
        path = sieve_to_cache(limit, force=args.force_build)

    print(f"Using mask: {path}")
    print("Opening mask (mmap)...")
    t0 = time.perf_counter()
    mv = open_mask(path)
    t1 = time.perf_counter()
    print(f"Opened in {t1 - t0:.3f}s. Limit={mv.limit:,}  bytes={len(mv):,}")

    # chunk selection
    env_chunk = os.environ.get("LOOKPRIME_CHUNK")
    env_chunk_val = int(env_chunk) if (env_chunk and env_chunk.isdigit()) else None

    if args.chunk is not None:
        chunk = int(args.chunk)
        chunk_note = ""
    elif env_chunk_val is not None:
        chunk = env_chunk_val
        chunk_note = " (env)"
    else:
        feats = detect_cpu_features()
        chunk = autotune_chunk(mask_bytes=len(mv), cpu_features=feats)
        chunk_note = " (auto)"

    print(f"Chunk: {chunk:,}{chunk_note}")
    print("Tip: set LOOKPRIME_CHUNK=65536 to change defaults globally.\n")

    results = []
    max_ns = []

    try:
        for i in range(iterations):
            c, maxn = count_primes_for_duration(mv, duration=duration, start_n=1, chunk=chunk)
            results.append(c)
            max_ns.append(maxn)

            if not args.quiet:
                rate = c / duration if duration > 0 else float("inf")
                print(f"Iter {i+1:>3}/{iterations}: hits={c:,}  hits/s={rate:,.0f}  max_n={maxn:,}")

        avg = statistics.mean(results)
        stdev = statistics.stdev(results) if len(results) > 1 else 0.0
        pse = 100 * ((stdev / math.sqrt(len(results))) / avg) if (len(results) > 1 and avg != 0) else 0.0
        avg_rate = avg / duration if duration > 0 else float("inf")

        print("\nSummary")
        print("-------")
        print(f"Mask file:          {path}")
        print(f"LIMIT:              {mv.limit:,}")
        print(f"DURATION:           {duration:.6f}s")
        print(f"ITERATIONS:         {iterations}")
        print(f"CHUNK:              {chunk}")
        print(f"Average hits:       {avg:,.2f}")
        print(f"Average hits/sec:   {avg_rate:,.2f}")
        print(f"Std dev:            {stdev:,.2f}")
        print(f"Percent Std Error:  {pse:.3f}%")
        print(f"Avg max n reached:  {int(statistics.mean(max_ns)):,}")
        return 0
    finally:
        mv.close()


def build_parser() -> argparse.ArgumentParser:
    p = argparse.ArgumentParser(prog="lookprime", description="lookprime CLI")
    sub = p.add_subparsers(dest="command", required=True)

    inf = sub.add_parser("info", help="Show CPU + cache + mask info")
    inf.set_defaults(func=cmd_info)

    cc = sub.add_parser("clear-cache", help="Delete cached mask files")
    cc.set_defaults(func=cmd_clear_cache)

    ip = sub.add_parser("isprime", help="Check primality of a number")
    ip.add_argument("n", type=int, help="Number to test")
    ip.add_argument("--limit", type=int, default=None, help="Ensure sieve covers at least this limit")
    ip.set_defaults(func=cmd_isprime)

    fac = sub.add_parser("factor", help="Factor an integer into prime powers")
    fac.add_argument("n", type=int, help="Integer to factor")
    fac.add_argument("--limit", type=int, default=None, help="Ensure sieve covers at least this limit")
    fac.set_defaults(func=cmd_factor)

    bld = sub.add_parser("build", help="Build and cache a prime lookup mask (bitset + mmap)")
    bld.add_argument("--limit", type=int, required=True, help="Max n supported by lookup")
    bld.add_argument("--output", type=str, default=None, help="Write to this path (default: cache)")
    bld.add_argument("--force", action="store_true", help="Rebuild even if output exists")
    bld.set_defaults(func=cmd_build)

    b = sub.add_parser("benchmark", help="Benchmark prime lookup speed (uses mmap cache)")
    b.add_argument(
        "--limit",
        type=int,
        default=DEFAULT_LIMIT,
        help="Limit (used for cache if --mask not set)",
    )
    b.add_argument("--mask", type=str, default=None, help="Use a specific mask file instead of cache")
    b.add_argument("--duration", type=float, default=1.0, help="Seconds per iteration")
    b.add_argument("--iterations", type=int, default=25, help="Number of iterations")
    b.add_argument("--chunk", type=int, default=None, help="Time-check granularity (auto if omitted)")
    b.add_argument("--quiet", action="store_true", help="Only print summary")
    b.add_argument("--force-build", action="store_true", help="Rebuild cache mask even if present")
    b.set_defaults(func=cmd_benchmark)

    return p


def main(argv=None) -> int:
    parser = build_parser()
    args = parser.parse_args(argv)
    return args.func(args)


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations

import mmap
import os
import struct
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Iterator, Optional

from .core import sieve_odd_mask

# Bitset file format:
# header:
#   magic   8 bytes: b"LPMASK02"
#   version u32     : 2
#   limit   u64     : max n supported
#   nbits   u64     : number of bits in bitset (== (limit//2)+1)
# data:
#   bitset bytes little-endian within each byte:
#     bit i corresponds to odd number (2*i + 1)
#     i=0 -> 1 (not prime)
#
MAGIC = b"LPMASK02"
VERSION = 2
HEADER_STRUCT = struct.Struct("<8sIQQ")
HEADER_SIZE = HEADER_STRUCT.size


def _platform() -> str:
    return sys.platform


def cache_dir() -> Path:
    """
    Auto-cache location:
      - Linux: $XDG_CACHE_HOME/lookprime or ~/.cache/lookprime
      - macOS: ~/Library/Caches/lookprime
      - Windows: %LOCALAPPDATA%\\lookprime\\Cache
    """
    if os.name == "nt":
        base = os.environ.get("LOCALAPPDATA")
        if base:
            return Path(base) / "lookprime" / "Cache"
        return Path.home() / "AppData" / "Local" / "lookprime" / "Cache"

    if _platform() == "darwin":
        return Path.home() / "Library" / "Caches" / "lookprime"

    xdg = os.environ.get("XDG_CACHE_HOME")
    if xdg:
        return Path(xdg) / "lookprime"
    return Path.home() / ".cache" / "lookprime"


def default_mask_path(limit: int, base_dir: Optional[Path] = None) -> Path:
    base = base_dir or cache_dir()
    return base / f"lookprime_mask_{limit}.lpm"


def _pack_odd_mask_to_bitset(mask01: bytearray) -> bytes:
    """
    Convert odd-only 0/1 byte mask to bitset bytes.
    bit i corresponds to mask01[i].
    """
    nbits = len(mask01)
    nbytes = (nbits + 7) // 8
    out = bytearray(nbytes)
    for i, v in enumerate(mask01):
        if v:
            out[i >> 3] |= 1 << (i & 7)
    return bytes(out)


@dataclass
class MaskView:
    """
    Memory-mapped bitset view.
    Index with i = n >> 1 (odd half-index) -> returns 0 or 1
    """
    limit: int
    nbits: int

    _file: object
    _mmap: mmap.mmap
    _buf: memoryview
    _data_off: int
    _nbytes: int

    def __len__(self) -> int:
        # number of bits is nbits; but len() used elsewhere as bytes proxy
        return self._nbytes

    def __getitem__(self, idx: int) -> int:
        # idx is half-index; return bit
        if idx < 0 or idx >= self.nbits:
            return 0
        b = self._buf[self._data_off + (idx >> 3)]
        return (b >> (idx & 7)) & 1

    def close(self) -> None:
        try:
            self._buf.release()
        except Exception:
            pass
        try:
            self._mmap.close()
        finally:
            try:
                self._file.close()
            except Exception:
                pass

    # ---------- Fast bit counting/scanning helpers ----------

    def _data_slice(self) -> memoryview:
        return self._buf[self._data_off : self._data_off + self._nbytes]

    def count_ones_upto(self, bit_idx_inclusive: int) -> int:
        """
        Count set bits in [0..bit_idx_inclusive] (inclusive).
        Uses int.from_bytes(...).bit_count() in chunks.
        """
        if bit_idx_inclusive < 0:
            return 0
        if bit_idx_inclusive >= self.nbits:
            bit_idx_inclusive = self.nbits - 1

        end_bit = bit_idx_inclusive
        end_byte = end_bit >> 3
        end_in_byte = end_bit & 7

        data = self._data_slice()

        # Full bytes before end_byte
        full = 0
        if end_byte > 0:
            # count bytes [0:end_byte)
            # chunk to avoid huge int allocations
            CH = 8192
            i = 0
            while i < end_byte:
                j = min(end_byte, i + CH)
                full += int.from_bytes(data[i:j], "little").bit_count()
                i = j

        # Partial last byte [end_byte]
        last = data[end_byte]
        mask = (1 << (end_in_byte + 1)) - 1
        full += (last & mask).bit_count()

        return full

    def iter_set_bits(self, start_bit: int, end_bit_exclusive: int) -> Iterator[int]:
        """
        Yield set bit indices in [start_bit, end_bit_exclusive).
        Efficient scanning by bytes; only visits nonzero bytes.
        """
        if start_bit < 0:
            start_bit = 0
        if end_bit_exclusive > self.nbits:
            end_bit_exclusive = self.nbits
        if start_bit >= end_bit_exclusive:
            return

        data = self._data_slice()
        start_byte = start_bit >> 3
        end_byte = (end_bit_exclusive + 7) >> 3

        # handle first byte with mask
        b = int(data[start_byte])
        first_mask = 0xFF & (~((1 << (start_bit & 7)) - 1))
        b &= first_mask

        byte_i = start_byte
        while byte_i < end_byte:
            if byte_i != start_byte:
                b = int(data[byte_i])

            if b:
                # enumerate set bits in this byte
                while b:
                    lsb = b & -b
                    bit_pos = (lsb.bit_length() - 1)
                    bit_idx = (byte_i << 3) + bit_pos
                    if bit_idx >= end_bit_exclusive:
                        return
                    yield bit_idx
                    b ^= lsb

            byte_i += 1
            if byte_i == end_byte:
                break

            # if last byte, apply end mask at end in loop top next cycle

        # If end_bit_exclusive not aligned, iter_set_bits may yield bits beyond.
        # We check bit_idx >= end_bit_exclusive above, so OK.


def open_mask(path: Path) -> MaskView:
    path = Path(path)
    f = path.open("rb")
    mm = mmap.mmap(f.fileno(), 0, access=mmap.ACCESS_READ)
    buf = memoryview(mm)

    if len(buf) < HEADER_SIZE:
        mm.close()
        f.close()
        raise ValueError("Invalid mask file: too small")

    magic, ver, limit, nbits = HEADER_STRUCT.unpack_from(buf, 0)
    if magic != MAGIC or ver != VERSION:
        mm.close()
        f.close()
        raise ValueError("Invalid/unsupported mask file")

    nbytes = (int(nbits) + 7) // 8
    data_off = HEADER_SIZE
    if data_off + nbytes > len(buf):
        mm.close()
        f.close()
        raise ValueError("Invalid mask file: truncated")

    return MaskView(
        limit=int(limit),
        nbits=int(nbits),
        _file=f,
        _mmap=mm,
        _buf=buf,
        _data_off=data_off,
        _nbytes=nbytes,
    )


def write_mask_file(path: Path, limit: int, bitset_bytes: bytes, nbits: int) -> None:
    path = Path(path)
    path.parent.mkdir(parents=True, exist_ok=True)

    tmp = path.with_suffix(path.suffix + ".tmp")
    header = HEADER_STRUCT.pack(MAGIC, VERSION, int(limit), int(nbits))

    with tmp.open("wb") as f:
        f.write(header)
        f.write(bitset_bytes)
        f.flush()
        os.fsync(f.fileno())

    tmp.replace(path)


def sieve_to_cache(limit: int, base_dir: Optional[Path] = None, force: bool = False) -> Path:
    """
    Build the sieve and store it in the cache (bitset format).
    If present and valid, reuse.
    """
    path = default_mask_path(limit, base_dir=base_dir)
    if path.exists() and not force:
        try:
            mv = open_mask(path)
            mv.close()
            return path
        except Exception:
            pass

    mask01 = sieve_odd_mask(limit)  # bytearray of 0/1 for odd indices
    bitset = _pack_odd_mask_to_bitset(mask01)
    nbits = len(mask01)
    write_mask_file(path, limit=limit, bitset_bytes=bitset, nbits=nbits)
    return path
